from math_j.functions import *
import itertools
import re
import operator


_T = typing.TypeVar("_T")


@std.functions.experimental_class
class Array(typing.Generic[_T]):
    """
    Class that creates an array with a size specified at initialization that supports standard mathematical operations.
    NOTE: While an array is hashable, it must be locked (immutable) or __hash__ will raise a MathJError
    """
    def __init__(self, *items: _T, lock=False):
        self._items = list(items)
        self.locked = lock

    # <editor-fold desc="Iterable and indexable methods">
    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)

    def __contains__(self, item: _T):
        return item in self._items

    def __getitem__(self, item: typing.SupportsIndex | slice) -> _T:
        return self._items[item]

    def __setitem__(self, index: typing.SupportsIndex | slice, o: _T) -> None:
        if self.locked:
            raise MathJError("Cannot modify locked Array")
        self._items[index] = o
    # </editor-fold>

    # <editor-fold desc="Arithmetic operators">
    def __arithmetic_template(self, op_name: str, mod: str = None):
        if op_name not in ("add", "sub", "mul", "truediv", "floordiv", "pow"):
            raise AttributeError("Operator names must be in the set of {'add', 'sub', 'mul', 'truediv', "
                                 "'floordiv', 'pow'}, found: " + op_name)
        if mod not in (None, "r", "i"):
            raise AttributeError("Operator modifiers must be in the set of {None, 'r', 'i'}, found: " + mod)
        op = getattr(operator, "__" + op_name + "__")
        cls = type(self)[_T]

        def template(other: tuple | list | typing.Iterable[_T]):
            other = tuple(other)
            # Standard arithmetic
            if mod is None:
                return std.functions.switch(len(other), {
                    1: cls(op(val, other[0]) for val in self),
                    len(self): cls[_T](op(val, other[i]) for i, val in enumerate(self))
                }, error=MathJError("Array arithmetic requires other to be a singleton, or an Iterable of length 1 or "
                                    "len(self)"))
            # Reversed arithmetic
            if mod == "r":
                return std.functions.switch(len(other), {
                    1: cls(op(other[0], val) for val in self),
                    len(self): cls(op(other[i], val) for i, val in enumerate(self))
                }, error=MathJError("Array arithmetic requires other to be a singleton, or an Iterable of length 1 or "
                                    "len(self)"))
            # In-place arithmetic
            if len(other) == 1:
                for i, val in enumerate(self):
                    self[i] = op(val, other[0])
            elif len(other) == len(self):
                for i, val in enumerate(self):
                    self[i] = op(val, other[i])
            else:
                raise MathJError("Array arithmetic requires other to be a singleton, or an Iterable of length 1 or "
                                "len(self)")
            return self

        return template

    def __neg__(self):
        return type(self)(*(-val for val in self))

    def __add__(self, other):
        return self.__arithmetic_template("add")(other)

    def __radd__(self, other):
        return self.__arithmetic_template("add", "r")(other)

    def __iadd__(self, other):
        return self.__arithmetic_template("add", "i")(other)

    def __sub__(self, other):
        return self.__arithmetic_template("sub")(other)

    def __rsub__(self, other):
        return self.__arithmetic_template("sub", "r")(other)

    def __isub__(self, other):
        return self.__arithmetic_template("sub", "i")(other)

    def __mul__(self, other):
        return self.__arithmetic_template("mul")(other)

    def __rmul__(self, other):
        return self.__arithmetic_template("mul", "r")(other)

    def __imul__(self, other):
        return self.__arithmetic_template("mul", "i")(other)

    def __truediv__(self, other):
        return self.__arithmetic_template("truediv")(other)

    def __rtruediv__(self, other):
        return self.__arithmetic_template("truediv", "r")(other)

    def __itruediv__(self, other):
        return self.__arithmetic_template("truediv", "i")(other)

    def __floordiv__(self, other):
        return self.__arithmetic_template("floordiv")(other)

    def __rfloordiv__(self, other):
        return self.__arithmetic_template("floordiv", "r")(other)

    def __ifloordiv__(self, other):
        return self.__arithmetic_template("floordiv", "i")(other)

    def __pow__(self, other, modulo=None):
        if not (isinstance(other, tuple) or isinstance(other, list)):
            other = tuple(other) if hasattr(other, "__iter__") else (other,)
        if len(other) == 1:
            other = other[0]
        is_iterable = hasattr(other, "__getitem__")
        return type(self)(*(pow(self[i], other[i] if is_iterable else other, modulo) for i in range(len(self))))

    def __rpow__(self, other):
        if not (isinstance(other, tuple) or isinstance(other, list)):
            other = tuple(other) if hasattr(other, "__iter__") else (other,)
        if len(other) == 1:
            other = other[0]
        is_iterable = hasattr(other, "__getitem__")
        return type(self)(*((other[i] if is_iterable else other)**self[i] for i in range(len(self))))

    def __ipow__(self, other):
        return self.__arithmetic_template("pow", "i")
    # </editor-fold>

    # <editor-fold desc="Comparators">
    def __eq__(self, other):
        if not (isinstance(other, tuple) or isinstance(other, list)):
            other = tuple(other) if hasattr(other, "__iter__") else (other,)
        if len(other) == 1:
            other = other[0]
        is_iterable = hasattr(other, "__getitem__")
        if is_iterable and not (hasattr(other, "__len__") and len(other) == len(self)):
            return False
        return all(self[i] == (other[i] if is_iterable else other) for i in range(len(self)))

    def __ne__(self, other):
        if not (isinstance(other, tuple) or isinstance(other, list)):
            other = tuple(other) if hasattr(other, "__iter__") else (other,)
        if len(other) == 1:
            other = other[0]
        is_iterable = hasattr(other, "__getitem__")
        if is_iterable and not (hasattr(other, "__len__") and len(other) == len(self)):
            return True
        return any(self[i] != (other[i] if is_iterable else other) for i in range(len(self)))
    # </editor-fold>

    # <editor-fold desc="Conversions">
    def __str__(self):
        return f"{type(self).__name__}({', '.join(f'{value}' for value in self)})"

    def __repr__(self):
        return f"{type(self).__name__}({', '.join(f'{value}: {type(value)}' for value in self)}, locked={self.locked})"
    # </editor-fold>

    def __hash__(self):
        if not self.locked:
            raise MathJError("Cannot hash standard_j.Array unless it is locked")
        return hash(tuple(self))

    # <editor-fold desc="Copy Methods">
    def __copy__(self):
        return type(self)(*self, lock=self.locked)

    def copy(self, lock: bool = None) -> typing.Self:
        return type(self)(*self, lock = self.locked if lock is None else lock)
    # </editor-fold>

    # <editor-fold desc="Mapping methods">
    def map(self, func: typing.Callable, *args, wrap_iter=False):
        if len(args) == 0:
            return type(self)[_T](*(func(value) for value in self))
        args = tuple(tuple(arg) if hasattr(arg, "__iter__") else tuple(arg for _ in range(len(self))) for arg in args)
        if any(len(arg) != len(self) for arg in args):
            raise IndexError("all args must have a length in (1, len(self))")
        new_items = []
        for index, value in enumerate(self):
            func_args = (value, *(arg[index] for arg in args))
            if wrap_iter:
                new_items.append(func(func_args))
            else:
                new_items.append(func(*func_args))
        return type(self)[_T](*new_items)

    def imap(self, func: typing.Callable, *args, wrap_iter=False):
        if self.locked:
            raise MathJError("Cannot modify immutable Array")
        if len(args) == 0:
            for index, value in enumerate(self):
                self[index] = func(value)
            return self
        args = tuple(tuple(arg) if hasattr(arg, "__iter__") else tuple(arg for _ in range(len(self))) for arg in args)
        if any(len(arg) != len(self) for arg in args):
            raise IndexError("all args must have a length in (1, len(self))")
        for index, value in enumerate(self):
            func_args = (value, *(arg[index] for arg in args))
            if wrap_iter:
                self[index] = func(func_args)
            else:
                self[index] = func(*func_args)
        return self

    def set(self, *other):
        return self.imap(lambda _, o: o, *other)
    # </editor-fold>


class Vector(Array):
    """
    Subclass of Array that provides a magnitude property and normalize function
    """
    @property
    def magnitude(self):
        return sum(item**2 for item in self)**0.5

    def normalize(self):
        """
        NOTE: Not recursive, thus magnitude after normalization may be "close" to 1
        :return: Vector
        """
        return self.__itruediv__(self.magnitude)


class Fraction(object):
    """
    Class that allows for fraction mathematical operations. Usually more accurate than float operations
    """
    def __init__(self, numerator: NumericType, denominator: NumericType = 1, *, expanded_display=False):
        if isinstance(numerator, Fraction):
            self._items = [numerator.numerator, numerator.denominator * denominator]
        else:
            self._items = [numerator, denominator]
        self.expanded_display = expanded_display
        self.simplify()

    # <editor-fold desc="Properties">
    @property
    def numerator(self):
        return self._items[0]

    @numerator.setter
    def numerator(self, value):
        self._items[0] = value
        self.simplify()

    @property
    def denominator(self):
        return self._items[1]

    @denominator.setter
    def denominator(self, value):
        if value == 0:
            raise ZeroDivisionError("Cannot have a zero denominator")
        self._items[1] = value
        self.simplify()
    # </editor-fold>

    # <editor-fold desc="Arithmetic operators">
    def __arithmetic_template(self, op_name: str, mod: str = None):
        if mod not in (None, "r", "i"):
            raise AttributeError("Operator modifiers must be in the set of {None, 'r', 'i'}, found: " + mod)
        op = getattr(operator, "__" + op_name + "__")
        cls = type(self)

        def add_sub_template(other):
            if not isinstance(other, Fraction):
                other = Fraction(other)
            if mod == "i":
                self._items[0] = op(self.numerator * other.denominator, other.numerator * self.denominator)
                self._items[1] *= other.denominator
                return self.simplify()
            if mod is None:
                return cls(op(self.numerator * other.denominator, other.numerator * self.denominator),
                           self.denominator * other.denominator, expanded_display=self.expanded_display)
            return cls(op(other.numerator * self.denominator, self.numerator * other.denominator),
                       self.denominator * other.denominator, expanded_display=self.expanded_display)

        def mul_template(other):
            if not isinstance(other, Fraction):
                other = Fraction(other)
            if mod == "i":
                self._items[0] *= other.numerator
                self._items[1] *= other.denominator
                return self.simplify()
            # mod in (None, "r")
            return cls(self.numerator * other.numerator, self.denominator * other.denominator,
                       expanded_display=self.expanded_display)

        def div_template(other):
            if not isinstance(other, Fraction):
                other = Fraction(1, other)
            else:
                other = other.reciprocal()
            return mul_template(other)

        if op_name in ("add", "sub"):
            return add_sub_template
        if op_name == "mul":
            return mul_template
        if op_name == "truediv":
            return div_template
        # Error handling
        raise AttributeError("Operator names must be in the set of {'add', 'sub', 'mul', 'truediv'}, found: " + op_name)
    
    def __neg__(self):
        return type(self)(-self.numerator, self.denominator)

    def __add__(self, other):
        return self.__arithmetic_template("add")(other)

    def __radd__(self, other):
        return self.__arithmetic_template("add", "r")(other)

    def __iadd__(self, other):
        return self.__arithmetic_template("add", "i")(other)

    def __sub__(self, other):
        return self.__arithmetic_template("sub")(other)

    def __rsub__(self, other):
        return self.__arithmetic_template("sub", "r")(other)

    def __isub__(self, other):
        return self.__arithmetic_template("sub", "i")(other)

    def __mul__(self, other):
        return self.__arithmetic_template("mul")(other)

    def __rmul__(self, other):
        return self.__arithmetic_template("mul", "r")(other)

    def __imul__(self, other):
        return self.__arithmetic_template("mul", "i")(other)

    def __truediv__(self, other):
        return self.__arithmetic_template("truediv")(other)

    def __rtruediv__(self, other):
        return self.__arithmetic_template("truediv", "r")(other)

    def __itruediv__(self, other):
        return self.__arithmetic_template("truediv", "i")(other)

    def __pow__(self, power, modulo=None):
        if power < 0:
            return type(self)(pow(self.denominator, abs(power), modulo), pow(self.numerator, abs(power), modulo))
        return type(self)(pow(self.numerator, power, modulo), pow(self.denominator, power, modulo))

    def __rpow__(self, other):
        return other**float(self)

    def __ipow__(self, other):
        if other < 0:
            other = -other
            return self.set(self.denominator ** other, self.numerator ** other)
        return self.set(self.numerator ** other, self.denominator ** other)
    # </editor-fold>

    # <editor-fold desc="Overridden comparators">
    def __comparator_template(self, comp_name: str):
        if comp_name not in ("eq", "ne", "lt", "le", "gt", "ge"):
            raise AttributeError("Operator names must be in the set of {'eq', 'ne', 'lt', 'le', 'gt', 'ge'}, "
                                 "found: " + comp_name)
        comp = getattr(operator, "__" + comp_name + "__")

        def template(other):
            if not isinstance(other, Fraction):
                other = Fraction(other)
            return comp(self.numerator * other.denominator, other.numerator * self.denominator)

        return template

    def __eq__(self, other):
        return self.__comparator_template("eq")(other)

    def __ne__(self, other):
        return self.__comparator_template("ne")(other)

    def __lt__(self, other):
        return self.__comparator_template("lt")(other)

    def __le__(self, other):
        return self.__comparator_template("le")(other)

    def __gt__(self, other):
        return self.__comparator_template("gt")(other)

    def __ge__(self, other):
        return self.__comparator_template("ge")(other)
    # </editor-fold>

    # <editor-fold desc="Conversions">
    def __float__(self):
        return float(self.numerator / self.denominator)

    def __trunc__(self):
        return float(self).__trunc__()

    def __int__(self):
        return int(float(self))

    def __str__(self):
        if self.expanded_display:
            num = f"{self.numerator}"
            den = f"{self.denominator}"
            new_len = max(len(num), len(den)) + 2
            return "\n".join((num.center(new_len, " "), "".join("-" for _ in range(new_len)), den.center(new_len, " ")))
        return f"{self.numerator}/{self.denominator}"

    def __repr__(self):
        return f"{type(self).__name__}(numerator: {self.numerator}, denominator: {self.denominator})"
    # </editor-fold>

    # <editor-fold desc="Copy methods">
    def __copy__(self):
        return type(self)(self.numerator, self.denominator, expanded_display=self.expanded_display)

    def copy(self, expanded_display: bool = None):
        return type(self)(self.numerator, self.denominator,
                          expanded_display = self.expanded_display if expanded_display is None else expanded_display)
    # </editor-fold>

    def reciprocal(self):
        return type(self)(self.denominator, self.numerator, expanded_display=self.expanded_display)

    @std.functions.recursion_guarded
    def simplify(self):
        contains_complex = isinstance(self.numerator, complex) or isinstance(self.denominator, complex)
        if contains_complex:
            # Convert to complex
            self.numerator += 0j
            self.denominator += 0j

            # if the denominator only consists of an imaginary part, move it to the numerator
            if self.denominator.real == 0 and self.denominator.imag != 0:
                self.numerator *= 1j
                self.denominator *= 1j

        # if the denominator is 0, raise ZeroDivisionError
        if self.denominator == 0:
            raise ZeroDivisionError("Cannot have a zero denominator")
        # if the numerator is 0, it always simplifies to 0/1
        if self.numerator == 0:
            self.numerator = 0
            self.denominator = 1
            return self

        # multiply by 10 until the numerator and denominator do not have floating points
        if contains_complex:
            while not (self.numerator.real % 1 == self.numerator.imag % 1 ==
                       self.denominator.real % 1 == self.denominator.imag % 1 == 0):
                self.numerator *= 10
                self.denominator *= 10
        else:
            while not (self.numerator % 1 == self.denominator % 1 == 0):
                self.numerator *= 10
                self.denominator *= 10

        # Divide by gcd to simplify fraction
        hcf = gcd(self.numerator, self.denominator)
        self.numerator /= hcf
        self.denominator /= hcf

        # if the denominator is negative move the negative to the numerator
        if contains_complex:
            if self.denominator.real < 0 or (self.denominator.real == 0 and self.denominator.imag < 0):
                self.numerator *= -1
                self.denominator *= -1
        else:
            if self.denominator < 0:
                self.numerator *= -1
                self.denominator *= -1

        if contains_complex:
            # Addresses glitch where 0+aj * -1 == -0-aj (Instead resulting in -aj)
            self.numerator *= 1
            self.denominator *= 1

            # Convert back to ints if imaginary parts are 0
            if self.numerator.imag == 0:
                self.numerator = int(self.numerator.real)
            if self.denominator.imag == 0:
                self.denominator = int(self.denominator.real)

        return self

    def set(self, numerator: NumericType, denominator: NumericType):
        self._items[0] = numerator
        self._items[1] = denominator
        return self.simplify()


@std.functions.experimental_class
class Polynomial(object):
    def __init__(self, *coefficients: RealNumericType, variable: str = "x", reverse_coefficients=False):
        """
        Class used for representing a polynomial of a single variable
        :param coefficients: [x^0: c1, x^1: c2, ..., x^n: cN] coefficients of the polynomial starting with
                             the constant term and increasing in power
        :param variable: str The variable that this is a polynomial of
        :param reverse_coefficients: bool used to reverse the order of the input coefficients
        """
        self._items = list(coefficients[::-1] if reverse_coefficients else coefficients)
        self.variable = variable

    @classmethod
    def from_str(cls, string: str):
        items: list[_T] = [0]
        string = string.replace("**", "^").replace(" ", "")
        variable = None

        def compute_term(expression: str):
            expression = expression.replace("**", "^").replace(" ", "")
            match_data = re.search(r"\D[^^]*", expression)
            if match_data is None:
                return None, eval(expression), 0
            else:
                return (match_data.group(),
                        eval(expression[:match_data.span()[0]]) if match_data.span()[0] > 0 else 0,
                        eval(expression[match_data.span()[1] + 1:]) if match_data.span()[1] < len(expression) else 1)

        while len(string) > 0:
            next_sep = max(string.rfind("+"), string.rfind("-"))
            coefficient, exponent, var = compute_term(string if next_sep == -1 else string[next_sep:])
            if var is not None:
                if variable is None:
                    variable = var
                elif variable != var:
                    raise MathJError("Polynomial class only supports single variable evaluation")
            while len(items) < exponent - 1:
                items.append(0)
            items[exponent] += coefficient
            string = string[:next_sep]

        return cls(items, variable=variable)

    # <editor-fold desc="Iterable and indexable methods">
    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)

    def __contains__(self, item: _T):
        return item in self._items

    def __getitem__(self, item: typing.SupportsIndex | slice) -> _T:
        return self._items[item]

    def __setitem__(self, index: typing.SupportsIndex | slice, o: _T) -> None:
        self._items[index] = o
    # </editor-fold>

    # <editor-fold desc="Arithmetic operators">
    def __arithmetic_template(self, op_name: str, mod: str = None):
        if mod not in (None, "r", "i"):
            AttributeError("Operator modifiers must be in the set of {None, 'r', 'i'}, found: " + mod)
        op = getattr(operator, "__" + op_name + "__")
        cls = type(self)

        def add_sub_template(other: Polynomial | RealNumericType):
            if isinstance(other, RealNumericType):
                if mod == "i":
                    self[0] = op(self[0], other)
                    return self
                return cls(*(op(self[0], other) if mod is None else op(other, self[0]), *self[1:]),
                           variable=self.variable)

            if isinstance(other, Polynomial):
                if self.variable == other.variable:
                    if mod is None:
                        return cls(*(op(self[i] if i < len(self) else 0, other[i] if i < len(other) else 0)
                                         for i in range(max(len(self), len(other)))), variable=self.variable)
                    # Mod is 'i', since Polynomial wouldn't call a reversed arithmetic with a subclass of itself
                    self.extend_to(len(other))
                    for i, val in enumerate(self):
                        self[i] = op(val, other[i])
                    return self
                # Error handling
                raise MathJError("Single variable polynomials must match variables for arithmetic operations")
            raise NotImplementedError(f"Polynomial only supports arithmetic operations with Polynomial "
                                      f"| int | float, found: {type(other)}")

        def mul_template(other: Polynomial | RealNumericType):
            if isinstance(other, RealNumericType):
                if mod == "i":
                    for i, val in enumerate(self):
                        self[i] = op(val, other)
                    return self
                return cls(*(op(val, other) if mod is None else op(other, val) for val in self), variable=self.variable)

            if isinstance(other, Polynomial):
                if self.variable == other.variable:
                    if mod is None:
                        output = cls(*(0 for _ in range(len(self) + len(other) - 1)))
                        for i, val_1 in enumerate(self):
                            for j, val_2 in enumerate(other):
                                output[i + j] += op(val_1, val_2)
                        return output
                    # Mod is 'i', since Polynomial wouldn't call a reversed arithmetic with a subclass of itself
                    temp = self.copy()
                    self.extend_to(len(self) + len(other))
                    for i in range(len(self)):
                        self[i] = 0
                    for i, val_1 in enumerate(temp):
                        for j, val_2 in enumerate(other):
                            self[i + j] += op(val_1, val_2)
                    return self
                # Error handling
                raise MathJError("Single variable polynomials must match variables for arithmetic operations")
            raise NotImplementedError(f"Polynomial only supports arithmetic operations with Polynomial "
                                      f"| int | float, found: {type(other)}")

        def div_template(other: Polynomial | RealNumericType):
            if isinstance(other, RealNumericType):
                if mod is None:
                    return cls(*(op(val, other) for val in self), variable=self.variable)
                if mod == "r":
                    raise NotImplementedError("Polynomial doesn't support reversed division")
                # Mod is 'i'
                for i, val in enumerate(self):
                    self[i] = op(val, other)
                return self

            raise NotImplementedError("Polynomial only implements division with "
                                      "type[int | float], found: {type(other)}")

        def pow_template(other: Polynomial | RealNumericType):
            if isinstance(other, int) and 0 <= other:
                if mod is None:
                    if other == 0:
                        return cls(1, variable=self.variable)
                    if other == 1:
                        return self.copy()
                    output = self.copy()
                    for _ in range(other - 1):
                        output *= self
                    return output
                if mod == "r":
                    raise NotImplementedError("Polynomial doesn't support reversed exponentiation")
                # Mod is 'i'
                if other == 0:
                    for i in range(len(self)):
                        self[i] = 0
                    self[0] = 1
                elif other > 1:
                    for i in range(other):
                        self.__imul__(self)
                return self

            raise NotImplementedError("Polynomial only implements exponentiation with positive ints, "
                                      "found: {type(other)}")

        if op_name in ("add", "sub"):
            return add_sub_template
        if op_name == "mul":
            return mul_template
        if op_name in ("truediv", "floordiv"):
            return div_template
        if op_name == "pow":
            return pow_template

        raise AttributeError("Operator names must be in the set of {'add', 'sub', 'mul', 'truediv', "
                             "'floordiv'}, found: " + op_name)

    def __neg__(self):
        return type(self)(*(-val for val in self), variable=self.variable)

    def __add__(self, other):
        return self.__arithmetic_template("add")(other)

    def __radd__(self, other):
        return self.__arithmetic_template("add", "r")(other)

    def __iadd__(self, other):
        return self.__arithmetic_template("add", "i")(other)

    def __sub__(self, other):
        return self.__arithmetic_template("sub")(other)

    def __rsub__(self, other):
        return self.__arithmetic_template("sub", "r")(other)

    def __isub__(self, other):
        return self.__arithmetic_template("sub", "i")(other)

    def __mul__(self, other):
        return self.__arithmetic_template("mul")(other)

    def __rmul__(self, other):
        return self.__arithmetic_template("mul", "r")(other)

    def __imul__(self, other):
        return self.__arithmetic_template("mul", "i")(other)

    def __truediv__(self, other):
        return self.__arithmetic_template("truediv")(other)

    def __rtruediv__(self, other):
        return self.__arithmetic_template("truediv", "r")(other)

    def __itruediv__(self, other):
        return self.__arithmetic_template("truediv", "i")(other)

    def __floordiv__(self, other):
        return self.__arithmetic_template("floordiv")(other)

    def __rfloordiv__(self, other):
        return self.__arithmetic_template("floordiv", "r")(other)

    def __ifloordiv__(self, other):
        return self.__arithmetic_template("floordiv", "i")(other)

    def __pow__(self, power, modulo=None):
        if modulo is not None:
            raise AttributeError("Polynomial does not support modulo operations")
        return self.__arithmetic_template("pow")(power)

    def __ipow__(self, other):
        return self.__arithmetic_template("pow", "i")(other)
    # </editor-fold>

    # <editor-fold desc="Comparators">
    def __eq__(self, other: typing.Self):
        if isinstance(other, RealNumericType):
            for i, val in enumerate(self):
                if i == 0:
                    if val != other:
                        return False
                else:
                    if val != 0:
                        return False
            return True
        if not isinstance(other, Polynomial):
            raise TypeError("Polynomial can only be compared to Type[Polynomials | int | float]")
        if self.variable != other.variable:
            return False
        if len(other) < len(self):
            for i in range(len(self)):
                if self[i] != (other[i] if i < len(other) else 0):
                    return False
        elif len(self) < len(other):
            for i in range(len(other)):
                if other[i] != (self[i] if i < len(self) else 0):
                    return False
        else:
            for i, val in enumerate(self):
                if other[i] != val:
                    return False
        return True

    def __ne__(self, other: typing.Self):
        return not self == other
    # </editor-fold>

    def __str__(self):
        output = ""
        for i in range(len(self) - 1, -1, -1):
            c = self[i]
            if c == 0:
                if i == 0:
                    output += " + 0"
                continue
            if len(output) != 0:
                output += " - " if c < 0 else " + "
            elif c < 0:
                output += "-"
            if not (c in (-1, 1) and 0 < i):
                output += f"{abs(c)}"
            if 0 < i:
                output += f"{self.variable}"
            if 1 < i:
                output += f"^{i}"
        return output if output != "" else "0"

    # <editor-fold desc="Copy methods">
    def __copy__(self):
        return type(self)(*self, variable=self.variable)

    def copy(self, variable=None):
        return type(self)(*self, variable = self.variable if variable is None else variable)
    # </editor-fold>

    def extend_to(self, length: int):
        while len(self) < length:
            self._items.append(0)


@std.functions.experimental_class
class MultivariablePolynomial(object):
    def __init__(self, coefficients: typing.Iterable[RealNumericType], **variables: typing.Iterable[RealNumericType]):
        """
        Class used for representing a polynomial of multiple variables
        :param coefficients: Iterable[c1, c2, ..., cN] coefficients of the polynomial
        :param **variable: KeywordVariable[v1: Iterable[e1, e2, ..., eN], v2: Iterable[...], ..., vN: Iterable[...]]
                           the variables present in the polynomial, with each index containing the exponent of that
                           variable in the corresponding term
        :EXAMPLE: mp1 = MultivariablePolynomial([1, -2, 3, -4, 5], x=[4, 3, 2, 1, 0], y=[0, 1, 2, 3, 4])
                  print(mp1) -> (x^4) - 2(x^3)(y) + 3(x^2)(y^2) - 4(x)(y^3) + 5(y^4)
        """
        self.coefficients = list(coefficients)
        self.variables = {}
        for var, exp_list in variables.items():
            self.variables[var] = list(exp_list)
        self._verify()

    def _verify(self):
        for val in self.coefficients:
            if not isinstance(val, RealNumericType):
                raise TypeError(f"Coefficients for a polynomial must be of type[int, float], found: {type(val)}")
        length = len(self.coefficients)
        for exp_list in self.variables.values():
            for val in exp_list:
                if not isinstance(val, RealNumericType):
                    raise TypeError(f"Exponents for a polynomial must be of type[int, float], found: {type(val)}")
            length = max(length, len(exp_list))
        while len(self.coefficients) < length:
            self.coefficients.append(0)
        for var in self.variables.keys():
            while len(self.variables) < length:
                self.variables[var].append(0)


MatrixLike = NumericType | typing.Iterable[NumericType | typing.Iterable[NumericType]]


@std.functions.experimental_class
class Matrix(typing.Generic[_T]):
    """
    Class that uses 2-dimensional Arrays to allow for storage and mathematical operations of a matrix
    NOTE: similar to standard_j.Array, Matrix is hashable only when it is locked (immutable)
    """
    @std.functions.overload
    def __init__(self, value: MatrixLike, *, lock: bool = False):
        if not hasattr(value, "__iter__"):
            self._items = Array((Array(value, lock=lock),), lock=lock)
        else:
            value = tuple(value)
            if not hasattr(value[0], "__iter__"):
                self._items = Array((Array(*value, lock=lock),), lock=lock)
            else:
                self._items = Array(*(Array(*row, lock=lock) for row in value), lock=lock)

    @std.functions.overload
    def __init__(self, num_rows, num_cols, init_value=Fraction(0), *, lock: bool = False):
        self._items = Array((Array((init_value for _ in range(num_cols)), lock=lock)
                             for __ in range(num_rows)), lock=lock)

    def __iter__(self):
        return iter(self._items)

    @typing.overload
    def __getitem__(self, item: typing.SupportsIndex | slice) -> Array[_T]: pass

    @typing.overload
    def __getitem__(self, item: tuple[typing.SupportsIndex | slice, typing.SupportsIndex | slice]) -> _T: pass

    def __getitem__(self, item):
        if isinstance(item, tuple) and len(item) == 2:
            return self._items[item[0]][item[1]]
        return self._items[item]

    def __setitem__(self, key: typing.SupportsIndex | slice | tuple, value: _T):
        if hasattr(key, "__len__") and len(key) == 2:
            self[key[0]][key[1]] = value
        else:
            self[key].set(value)

    def __contains__(self, item):
        return item in itertools.chain(Array.__iter__(self._items))

    def __neg__(self):
        return type(self)((Array(-self[i, j] for j in range(self.size[1])) for i in range(self.size[0])))

    def __add__(self, other: MatrixLike | typing.Self):
        if not isinstance(other, Matrix):
            return type(self)((Array(self[i, j] + other for j in range(self.size[1])) for i in range(self.size[0])))
        if self.size != other.size:
            raise MathJError("Matrix addition only works with Matrices with matching sizes")
        return type(self)((Array(self[i, j] + other[i, j] for j in range(self.size[1])) for i in range(self.size[0])))

    def __radd__(self, other: MatrixLike):
        return self + other

    def __iadd__(self, other: MatrixLike | typing.Self):
        if isinstance(other, Matrix) and self.size != other.size:
            raise MathJError("Matrix addition only works with Matrices with matching sizes")
        value = other
        for i in range(self.size[0]):
            for j in range(self.size[1]):
                if isinstance(other, Matrix):
                    value = other[i, j]
                self[i, j] += value
        return self

    def __sub__(self, other: MatrixLike | typing.Self):
        if not isinstance(other, Matrix):
            return type(self)((Array(self[i, j] - other for j in range(self.size[1])) for i in range(self.size[0])))
        if self.size != other.size:
            raise MathJError("Matrix subtraction only works with Matrices with matching sizes")
        return type(self)((Array(self[i, j] - other[i, j] for j in range(self.size[1])) for i in range(self.size[0])))

    def __rsub__(self, other: MatrixLike):
        if not isinstance(other, Matrix):
            return type(self)((Array(other - self[i, j] for j in range(self.size[1])) for i in range(self.size[0])))
        if self.size != other.size:
            raise MathJError("Matrix subtraction only works with Matrices with matching sizes")
        return type(self)((Array(other[i, j] - self[i, j] for j in range(self.size[1])) for i in range(self.size[0])))

    def __isub__(self, other: MatrixLike | typing.Self):
        if isinstance(other, Matrix) and self.size != other.size:
            raise MathJError("Matrix subtraction only works with Matrices with matching sizes")
        value = other
        for i in range(self.size[0]):
            for j in range(self.size[1]):
                if isinstance(other, Matrix):
                    value = other[i, j]
                self[i, j] -= value
        return self

    def dot(self, other: MatrixLike | typing.Self):
        if not isinstance(other, Matrix):
            if not hasattr(other, "__iter__"):
                output = self.copy()
                for i in range(output.size[0]):
                    for j in range(output.size[1]):
                        output[i, j] *= other
                return output
            other = Matrix(other, lock=True)
        other: Matrix
        if self.size[1] != other.size[0]:
            raise MathJError("Matrix multiplication only works with Matrices with matching internal dimensions")
        output = Matrix(*self.size)
        for i in range(output.size[0]):
            for j in range(output.size[1]):
                row, col = self.row(i), other.column(j)
                output[i, j] = sum(row[0, index] * col[index, 0] for index in range(self.size[1]))
        return output

    __mul__ = dot

    def __rmul__(self, other: MatrixLike):
        if not isinstance(other, Matrix):
            return self * other
        return Matrix(other) * self

    def __imul__(self, other: NumericType):
        if isinstance(other, Matrix):
            raise MathJError("Matrix.__imul__ only works with scaler other")
        for i in range(self.size[0]):
            for j in range(self.size[1]):
                self[i, j] *= other
        return self

    def __truediv__(self, other: MatrixLike | typing.Self | NumericType):
        return self * other**(-1)

    def __rtruediv__(self, other: MatrixLike):
        return other * self**(-1)

    def __itruediv__(self, other: NumericType):
        if isinstance(other, int) or isinstance(other, float) or isinstance(other, complex):
            return self * other**(-1)
        raise MathJError("Matrix.__itruediv__ only works with scaler other")

    __floordiv__ = NotImplemented
    __rfloordiv__ = NotImplemented
    __ifloordiv__ = NotImplemented

    def __pow__(self, other: int, modulo=None) -> typing.Self:
        if not (isinstance(other, int) and -1 <= other):
            raise MathJError("Matrix can only be raised to a positive int or -1 (to invert)")
        if not self.is_square():
            raise MathJError("Matrix must be square to use __pow__")
        if other == -1:
            if self.determinant == 0:
                raise MathJError("Cannot invert singular matrix (det == 0)")
            return self.adjoint().transpose() / self.determinant
        output = self.copy()
        for i in range(other - 1):
            output *= self
        return output

    def __ipow__(self, other: int, modulo=None):
        if not (isinstance(other, int) and -1 <= other):
            raise MathJError("Matrix can only be raised to a positive int or -1 (to invert)")
        if not self.is_square():
            raise MathJError("Matrix must be square to use __pow__")
        if other == -1:
            if self.determinant == 0:
                raise MathJError("Cannot invert singular matrix (det == 0)")
            return self.set(self.adjoint().transpose() / self.determinant)
        for i in range(other - 1):
            self.__imul__(self)
        return self

    def __eq__(self, other: MatrixLike | typing.Self):
        if not isinstance(other, Matrix):
            return all(self[i, j] == other for i in range(self.size[0]) for j in range(self.size[1]))
        if other.size[0] == 1:
            if other.size[1] != self.size[1]:
                raise MathJError("Invalid number of columns for Matrix comparison")
            return all(self.row(i) == other.row(0) for i in range(self.size[1]))
        if other.size[1] == 1:
            if other.size[0] != self.size[0]:
                raise MathJError("Invalid number of rows for Matrix comparison")
            return all(self.column(i) == other.column(0) for i in range(self.size[0]))
        if other.size != self.size:
            raise MathJError("Invalid size for Matrix comparison")
        return all(self[i, j] == other[i, j] for i in range(self.size[0]) for j in range(self.size[1]))

    def __ne__(self, other):
        if not isinstance(other, Matrix):
            return any(self[i, j] != other for i in range(self.size[0]) for j in range(self.size[1]))
        if other.size[0] == 1:
            if other.size[1] != self.size[1]:
                raise MathJError("Invalid number of columns for Matrix comparison")
            return any(self[i] != other[0] for i in range(self.size[1]))
        if other.size[1] == 1:
            if other.size[0] != self.size[0]:
                raise MathJError("Invalid number of rows for Matrix comparison")
            return any(self.column(i) != other.column(i) for i in range(self.size[0]))
        if other.size != self.size:
            raise MathJError("Invalid size for Matrix comparison")
        return any(self[i, j] != other[i, j] for i in range(self.size[0]) for j in range(self.size[1]))

    def __hash__(self):
        return hash(self._items)

    def __copy__(self):
        return type(self)(self, lock=self._items.locked)

    def __str__(self):
        name = type(self).__name__
        col_lens = [max(len(str(val[0])) for val in self.column(i)) for i in range(self.size[1])]
        output = f"{name}("
        for i, row in enumerate(self._items):
            for col_num in range(len(row)):
                output += str(row[col_num]).rjust(col_lens[col_num], ' ') + ' '
            if i != self.size[0] - 1:
                output += "\n" + ' '.join('' for _ in range(len(name) + 1)) + ' '
        return output[:-1] + ')'

    def __repr__(self):
        output = f"{type(self).__name__}("
        output += "; ".join(', '.join(f'{value}: {type(value)}' for value in row) for row in self)
        return output + f"; locked={self._items.locked})"

    def copy(self, lock=False) -> typing.Self:
        return type(self)(self, lock=lock)

    def map(self, func: typing.Callable, *args, wrap_iter=False) -> typing.Self:
        output = self.copy()
        if len(args) == 0:
            for i in range(output.size[0]):
                for j in range(output.size[1]):
                    output[i, j] = func(output[i, j])
            return output
        args = list(args)
        for index, arg in enumerate(args):
            if not isinstance(arg, list):
                arg = list(arg)
            if len(arg) == 1:
                arg = list(arg for _ in range(output.size[0]))
            for row_index, row in enumerate(arg):
                if not isinstance(row, list):
                    row = list(row)
                if len(row) == 1:
                    row = list(row for _ in range(output.size[1]))
                arg[row_index] = row
            args[index] = arg
        if any(len(arg) != output.size[0] or any(len(row) != output.size[1] for row in arg) for arg in args):
            raise IndexError("all args must have a length in (1, len(self), self.size)")
        for i in range(output.size[0]):
            for j, value in enumerate(output[i]):
                func_args = (value, *(arg[i][j] for arg in args))
                if wrap_iter:
                    output[i, j] = func(func_args)
                else:
                    output[i, j] = func(*func_args)
        return output

    def imap(self, func: typing.Callable, *args, wrap_iter=False) -> typing.Self:
        if self._items.locked:
            raise MathJError("Cannot modify immutable Array")
        if len(args) == 0:
            for i in range(self.size[0]):
                for j in range(self.size[1]):
                    self[i, j] = func(self[i, j])
            return self
        args = list(args)
        for index, arg in enumerate(args):
            if not isinstance(arg, list):
                arg = list(arg)
            if len(arg) == 1:
                arg = list(arg for _ in range(self.size[0]))
            for row_index, row in enumerate(arg):
                if not isinstance(row, list):
                    row = list(row)
                if len(row) == 1:
                    row = list(row for _ in range(self.size[1]))
                arg[row_index] = row
            args[index] = arg
        if any(len(arg) != self.size[0] or any(len(row) != self.size[1] for row in arg) for arg in args):
            raise IndexError("all args must have a length in (1, len(self), self.size)")
        for i in range(self.size[0]):
            for j, value in enumerate(self[i]):
                func_args = (value, *(arg[i][j] for arg in args))
                if wrap_iter:
                    self[i, j] = func(func_args)
                else:
                    self[i, j] = func(*func_args)
        return self

    def set(self, *other: MatrixLike | typing.Self) -> typing.Self:
        return self.imap(lambda row, o: row.set(o), *other)

    @property
    def size(self) -> tuple[int, int]:
        num_rows = len(self._items)
        return num_rows, (len(self[0]) if num_rows > 0 else 0)

    def row(self, index: typing.SupportsIndex) -> typing.Self:
        return Matrix(self[index])

    def column(self, index: typing.SupportsIndex) -> typing.Self:
        return Matrix([self[i, index]] for i in range(self.size[0]))

    @property
    def inverse(self) -> typing.Self:
        return self**(-1)

    def is_square(self) -> bool:
        return self.size[0] == self.size[1]

    @property
    def determinant(self) -> NumericType:
        if not self.is_square():
            raise MathJError("Matrix must be square to find determinant")
        if self.size[0] == 1:
            return self[0, 0]
        if self.size[0] == 2:
            return (self[0, 0] * self[1, 1]) - (self[0, 1] * self[1, 0])
        return sum(self[i, 0] * self.get_cofactor(i, 0) for i in range(self.size[0]))

    def get_minor(self, i: int, j: int) -> NumericType:
        sub_matrix = Matrix(self.size[0] - 1, self.size[1] - 1)
        for r in range(self.size[0]):
            if r == i:
                continue
            for c in range(self.size[1]):
                if c == j:
                    continue
                sub_matrix[r - (0 if r < i else 1), c - (0 if c < j else 1)] = self[r, c]
        return sub_matrix.determinant

    def get_cofactor(self, i: int, j: int) -> NumericType:
        return (-1)**(i + j) * self.get_minor(i, j)

    def adjoint(self) -> typing.Self:
        adjoint = Matrix(*self.size)
        for i in range(self.size[0]):
            for j in range(self.size[1]):
                adjoint[i, j] = self.get_cofactor(i, j)
        return adjoint

    def transpose(self) -> typing.Self:
        output = Matrix(*self.size[::-1])
        for i in range(self.size[0]):
            for j in range(self.size[1]):
                output[j, i] = self[i, j]
        return output

    @classmethod
    def Identity(cls, size):
        output = cls(size, size)
        for i in range(size):
            output[i, i] = 1
        return output
